/**
 * @file basic_threads.cpp
 * @author Fabian Kantereit
 *
 * This file demonstrates the basic usage of threads.
 */

#include <thread>
#include <chrono>
#include <iostream>

void foo() {
  std::cout << "Thread " << std::this_thread::get_id() << std::endl;
}

int main(int argc, char *argv[]) {
  // determine current thread id
  std::cout << "main thread " << std::this_thread::get_id() << std::endl;

  // number of hardware threads
  std::cout << "This machine supports up to " << std::thread::hardware_concurrency()
            << " hardware threads" << std::endl;

  /**********************************
   * running threads
   **********************************/

  // thread in "not a thread" state
  std::thread t1;

  // thread is started immediately
  std::thread t2(foo);

  // run a lambda expression in our thread
  t1 = std::thread([](void) {
    std::cout << "Thread " << std::this_thread::get_id() << " called lambda"
              << std::endl;
    std::this_thread::sleep_for(std::chrono::seconds(5));
  });

  /**********************************
   * joining threads
   **********************************/

  std::cout << "T2 joinable: " << t2.joinable() << std::endl;

  // join the thread. Blocks until execution is completed
  t2.join();
  std::cout << "T2 joinable: " << t2.joinable() << std::endl;
  std::cout << "Thread 2 joined" << std::endl;

  /**********************************
   * detaching threads
   **********************************/

  std::cout << "T1 joinable: " << t1.joinable() << std::endl;
  t1.detach();
  std::cout << "Thread 1 detached" << std::endl;
  std::cout << "T1 joinable: " << t1.joinable() << std::endl;
  // the exact behaviour on termination is undefined if
  // detached threads are still running
  // despite the sleep above, the thread is terminated
  return 0;
}
