/**
 * @file additional_concepts
 * @author Fabian Kantereit
 */

#include <string>
#include <iostream>
#include <thread>
#include <mutex>

std::once_flag method_flag;

thread_local int number;


void once_print() {
  std::cout << "Hello World" << std::endl;
}

void foo() {
  std::call_once(method_flag, once_print);
}

void modify_number() {
  number = 6 ;
  std::cout << "Number in t3: " << number << std::endl;
}

int main(int argc, char *argv[])
{
  /******************************************
   * run foo in 2 threads
   * Hello World will be printed only once */

  std::thread t1(foo);
  std::thread t2(foo);

  t1.join();
  t2.join();

  /**
   * thread_local storage will be restricted to each thread*/
  number = 5;
  std::cout << "Number in main thread: " << number << std::endl;
  std::thread t3(modify_number);
  t3.join();
  std::cout << "Number in main thread: " << number << std::endl;

  return 0;
}
