/**
 * @file locks.cpp
 * @author Fabian Kantereit
 */

#include <iostream>
#include <thread>
#include <mutex>

std::mutex m_cout;

void safe_cout(const char *message) {
  try {
    std::lock_guard<std::mutex> lock(m_cout);
    std::cout << message << std::endl;

    static int count = 0;
    ++count;
    if (count == 1) throw 0;
  } catch (...) {
  }
}

void safe_cout_v2(const char *message) {
  try {
    std::unique_lock<std::mutex> lock(m_cout);
    std::cout << message << std::endl;

    static int count = 0;
    ++count;
    if (count == 1) throw 0;
    lock.unlock(); // can be explicitly unlocked
  } catch (...) {
  }
}

int main(int argc, char *argv[]) {
  std::thread t1(safe_cout, "Hello ");
  std::thread t2(safe_cout, "World ");
  std::thread t3(safe_cout, "Today");
  std::thread t4(safe_cout, "is a good day");

  t1.join();
  t2.join();
  t3.join();
  t4.join();


  t1 = std::thread(safe_cout, "Hello ");
  t2 = std::thread(safe_cout, "World ");
  t3 = std::thread(safe_cout, "Today");
  t4 = std::thread(safe_cout, "is a good day");

  t1.join();
  t2.join();
  t3.join();
  t4.join();

  return 0;
}
