/**
 * @file 07_condition_variable.cpp
 * @author Fabian Kantereit
 */

#include <iostream>
#include <thread>
#include <mutex>
#include <condition_variable>
#include <queue>
#include <chrono>

int n = 0;
std::queue<int> values;
std::mutex m;
std::condition_variable cv_value_ready;
bool done = false;
bool notified = false;

void produce_values() {
  for (int i = 0; i < 5; ++i) {
    std::this_thread::sleep_for(std::chrono::seconds(1));
    std::unique_lock<std::mutex> lock(m);
    std::cout << "producing " << i << '\n';
    values.push(i);

    // notify the consumer
    notified = true;
    cv_value_ready.notify_one();
  }

  // signal done
  std::unique_lock<std::mutex> lock(m);
  done = true;
  cv_value_ready.notify_one();
}

void consume_values() {
  std::unique_lock<std::mutex> lock(m);
  while (!done) {
    // wait for lock and notified, lock will be unlocked here
    cv_value_ready.wait(lock, [&]() { return notified || done; });
    // lock is reacquired here

    while (!values.empty()) {
      std::cout << "consuming " << values.front() << '\n';
      values.pop();
    }
    notified = false;
  }
}

int main(int argc, char *argv[])
{
  std::thread producer(produce_values);
  std::thread consumer(consume_values);

  producer.join();
  consumer.join();
}
