/**
 * @file deadlock.cpp
 * @author Fabian Kantereit
 */

#include <thread>
#include <mutex>
#include <iostream>
#include <chrono>

// comment out for better version
#define V1

int main(int argc, char *argv[])
{
  std::mutex m_a;
  std::mutex m_b;

#ifdef V1
  // Fehlerhafte Variante, hier kann ein deadlock auftreten
  std::thread t1([&](){
      std::lock_guard<std::mutex> lock_a(m_a);
      std::this_thread::sleep_for(std::chrono::milliseconds(50));
      std::lock_guard<std::mutex> lock_b(m_b);

      std::cout << "Hello World";
    });

  std::thread t2([&](){
      std::lock_guard<std::mutex> lock_b(m_b);
      std::lock_guard<std::mutex> lock_a(m_a);
      std::cout << "Hello World";
    });
#endif

#ifndef V1
  // bessere Variante
    std::thread t1([&](){
        std::lock(m_a, m_b);
        std::lock_guard<std::mutex> lock_a(m_a, std::adopt_lock);
        std::this_thread::sleep_for(std::chrono::milliseconds(50));
        std::lock_guard<std::mutex> lock_b(m_b, std::adopt_lock);

        std::cout << "Hello World" << std::endl;
      });

    std::thread t2([&](){
        std::lock(m_a, m_b);
        std::lock_guard<std::mutex> lock_b(m_b, std::adopt_lock);
        std::lock_guard<std::mutex> lock_a(m_a, std::adopt_lock);
        std::cout << "Hello World" << std::endl;
      });
#endif

  t1.join();
  t2.join();
}
