/**
 * @file multiThreadedException01.cpp
 * @author fabian
 *
 * Created on April 3, 2014, 11:40 AM
 */

#include <future>

#define BOOST_LOG_DYN_LINK
#include <boost/log/trivial.hpp>

/**
 * return exceptions with a exception_ptr
 * @param exc_thrown
 */
void threaded_exception(std::exception_ptr &exc_thrown) {
  try {
    // do critical stuff here
    throw std::exception();
    // throw 0;
    // int i = 1 / 0;
  } catch (...) {
    exc_thrown = std::current_exception();
    return;
  }
}

/**
 * demonstrate exception ptr
 */
void exception_example(void) {
  std::exception_ptr thrown_exc;

  std::thread t(threaded_exception, std::ref(thrown_exc));
  t.join();
  try {
    if (thrown_exc) {
      std::rethrow_exception(thrown_exc);
    }
  } catch (std::exception &e) {
    BOOST_LOG_TRIVIAL(error) << "\nError: " << e.what();
  } catch (...) {
    BOOST_LOG_TRIVIAL(error) << "\nGeneric Error";
  }
}

void promise_store_exception(void) {
  // use futures / promises to store exception
  // this allows to throw exceptions for each result the thread
  // useful if a thread is working on many operation and can continue
  std::promise<int> result;

  std::thread([&](void) {
                try {
                  throw std::exception();
                } catch (...) {
                  result.set_exception(std::current_exception());
                }
              }).detach();

  try {
    BOOST_LOG_TRIVIAL(debug) << result.get_future().get();
  } catch (std::exception &e) {
    BOOST_LOG_TRIVIAL(error) << "\nStored Exception: " << e.what();
  }

  // std::async will store excpetions in the future
  std::future<void> result2 = std::async([] { throw std::exception(); });
  try {
    result2.get();
  } catch (std::exception &e) {
    BOOST_LOG_TRIVIAL(error) << "\nAutostore: " << e.what();
  }
}

/*
 *
 */
int main(int argc, char **argv) {
  // exception handling in threads
  exception_example();

  // store exception in promise / future
  promise_store_exception();
  return 0;
}
